@@ -2,7 +2,7 @@
 //#![forbid(unsafe_code, unused_variables, unused_imports)]
 //#![no_std] only blocker in auto_impl check: https://github.com/bluealloy/revm/issues/4
 
-mod db;
+pub mod db;
 mod evm;
 mod evm_impl;
 mod inspector;
@@ -13,21 +13,21 @@ mod spec;
 mod subroutine;
 mod util;
 
-use evm_impl::Handler;
+pub use evm_impl::{EVMData, Host};
 
 pub type DummyStateDB = InMemoryDB;
 
 pub use db::{Database, DatabaseCommit, InMemoryDB};
 pub use evm::{new, EVM};
-pub use inspector::{Inspector, NoOpInspector};
+pub use inspector::{Inspector, NoOpInspector, OverrideSpec};
 pub use instructions::{
-    opcode::{self, OpCode, OPCODE_JUMPMAP},
+    opcode::{self, spec_opcode_gas, OpCode, OPCODE_JUMPMAP},
     Return,
 };
-pub use machine::Machine;
+pub use machine::{Gas, Machine};
 pub use models::*;
 pub use spec::*;
-pub use subroutine::Account;
+pub use subroutine::{Account, SubRoutine};
 
 extern crate alloc;
 
