@@ -1,8 +1,7 @@
 use crate::{models::SelfDestructResult, Return, KECCAK_EMPTY};
 use alloc::{vec, vec::Vec};
-use hashbrown::{hash_map::Entry, HashMap as Map};
-
 use core::mem::{self};
+use hashbrown::{hash_map::Entry, HashMap as Map};
 
 use bytes::Bytes;
 use primitive_types::{H160, H256, U256};
@@ -68,12 +67,30 @@ pub enum SlotChangeLog {
 pub struct SubRoutineCheckpoint {
     log_i: usize,
     changelog_i: usize,
-    depth: usize,
+}
+
+impl Default for SubRoutine {
+    fn default() -> Self {
+        Self::new()
+    }
 }
 
 impl SubRoutine {
-    pub fn new(precompiles: Map<H160, AccountInfo>) -> SubRoutine {
-        let state = precompiles
+    pub fn new() -> SubRoutine {
+        Self {
+            state: Map::new(),
+            logs: Vec::new(),
+            changelog: vec![Map::new(); 1],
+            depth: 0,
+        }
+    }
+
+    pub fn state(&mut self) -> &mut State {
+        &mut self.state
+    }
+
+    pub fn load_precompiles(&mut self, precompiles: Map<H160, AccountInfo>) {
+        let state: Map<H160, Account> = precompiles
             .into_iter()
             .map(|(k, value)| {
                 let mut acc = Account::from(value);
@@ -81,13 +98,19 @@ impl SubRoutine {
                 (k, acc)
             })
             .collect();
+        self.state.extend(state);
+    }
 
-        Self {
-            state,
-            logs: Vec::new(),
-            changelog: vec![Map::new(); 1],
-            depth: 0,
-        }
+    pub fn load_precompiles_default(&mut self, precompiles: &[H160]) {
+        let state: State = precompiles
+            .iter()
+            .map(|&k| {
+                let mut acc = Account::from(AccountInfo::default());
+                acc.filth = Filth::Precompile(false);
+                (k, acc)
+            })
+            .collect();
+        self.state.extend(state);
     }
 
     /// do cleanup and return modified state
@@ -378,7 +401,6 @@ impl SubRoutine {
         let checkpoint = SubRoutineCheckpoint {
             log_i: self.logs.len(),
             changelog_i: self.changelog.len(),
-            depth: self.depth,
         };
         self.depth += 1;
         self.changelog.push(Default::default());
@@ -456,10 +478,10 @@ impl SubRoutine {
     }
 
     /// load account into memory. return if it is cold or hot accessed
-    pub fn load_account<'a, DB: Database>(&'a mut self, address: H160, db: &mut DB) -> bool {
+    pub fn load_account<DB: Database>(&mut self, address: H160, db: &mut DB) -> bool {
         let is_cold = match self.state.entry(address) {
-            Entry::Occupied::<'a>(ref mut _entry) => false,
-            Entry::Vacant::<'a>(vac) => {
+            Entry::Occupied(ref mut _entry) => false,
+            Entry::Vacant(vac) => {
                 let acc: Account = db.basic(address).into();
                 // insert none in changelog that represent that we just loaded this acc in this subroutine
                 self.changelog
@@ -636,6 +658,10 @@ impl Filth {
     pub fn abandon_old_storage(&self) -> bool {
         matches!(self, Self::Destroyed | Self::NewlyCreated)
     }
+
+    pub fn is_precompile(&self) -> bool {
+        matches!(self, Self::Precompile(_))
+    }
     /// insert into dirty flag and return if slot was already dirty or not.
     #[inline]
     pub fn insert_dirty_original(&mut self, index: U256, present_value: U256) {
